# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Natively supported multimodal datasets."""

from streaming.multimodal.webvid import StreamingInsideWebVid as StreamingInsideWebVid
from streaming.multimodal.webvid import StreamingOutsideDTWebVid as StreamingOutsideDTWebVid
from streaming.multimodal.webvid import StreamingOutsideGIWebVid as StreamingOutsideGIWebVid
