# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Base module for dataset reader and writer."""

from streaming.base.format.base.reader import FileInfo, Reader

__all__ = ['FileInfo', 'Reader']
